"""
=====================================
FORMAL TESTS ON STOCHASTICITY SOURCES
=====================================
"""
from .shared import *
from numpy import sqrt
import scipy.stats

source = sp.source
wiener_source = sp.wiener_source
poisson_source = sp.poisson_source
cpoisson_source = sp.cpoisson_source

odd_wiener_source = sp.odd_wiener_source
even_poisson_source = sp.even_poisson_source
even_cpoisson_source = sp.even_cpoisson_source

true_wiener_source = sp.true_wiener_source
true_poisson_source = sp.true_poisson_source
true_cpoisson_source = sp.true_cpoisson_source

dw = sp.dw
odd_dw = sp.odd_dw
true_dw = sp.true_dw
dn = sp.dn
even_dn = sp.even_dn
true_dn = sp.true_dn
dj = sp.dj
even_dj = sp.even_dj
true_dj = sp.true_dj

norm_rv = sp.norm_rv
uniform_rv = sp.uniform_rv
exp_rv = sp.exp_rv
double_exp_rv = sp.double_exp_rv

true_source = sp.true_source
true_wiener_source = sp.true_wiener_source


# ---------------------
# sources general tests
# ---------------------

# main test
def test_source_general():
    np.random.seed(SEED)

    # do cases
    paths = [10]
    float_dtype = [None, np.float32]
    int_dtype = [int, np.int16]

    source_and_params = [
        (wiener_source, dict(vshape=())),
        (wiener_source, dict(vshape=(2, 3))),
        (wiener_source, dict(vshape=(2,), corr=((1., -.5), (-.5, 1.)))),
        (wiener_source, dict(vshape=(3, 2,), corr=((1., -.5), (-.5, 1.)))),
        (true_wiener_source, dict(vshape=())),
        (true_wiener_source, dict(vshape=(2, 3))),
        (true_wiener_source, dict(vshape=(2,), corr=((1., -.5), (-.5, 1.)))),
        (true_wiener_source, dict(vshape=(3, 2,),
                                  corr=((1., -.5), (-.5, 1.)))),
        (odd_wiener_source, dict(vshape=())),
        (odd_wiener_source, dict(vshape=(2, 3))),
        (odd_wiener_source, dict(vshape=(2,), corr=((1., -.5), (-.5, 1.)))),
        (odd_wiener_source, dict(vshape=(3, 2,), corr=((1., -.5), (-.5, 1.)))),
        (poisson_source, dict(vshape=())),
        (poisson_source, dict(vshape=(2,), lam=1.)),
        (poisson_source, dict(vshape=(2,), lam=((1.,), (2.,)))),
        (poisson_source, dict(vshape=(3, 2), lam=((1.,), (2.,)))),
        (poisson_source, dict(vshape=(2, 3, 5),
                              lam=np.arange(15.).reshape(3, 5, 1))),
        (cpoisson_source, dict(vshape=())),
        (cpoisson_source, dict(vshape=(3, 2), lam=((1.,), (2.,)))),
        (cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=norm_rv(a=1., b=5.), ptype=np.int16)),
        (cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=uniform_rv(a=1., b=5.))),
        (cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=double_exp_rv(pa=.3, a=1., b=5.))),
        (cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=scipy.stats.trapz(c=0.2, d=0.4))),
        (cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=scipy.stats.expon(scale=2.))),
        (true_cpoisson_source, dict(vshape=())),
        (true_cpoisson_source, dict(vshape=(3, 2), lam=((1.,), (2.,)))),
        (true_cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=norm_rv(a=1., b=5.))),
        (true_cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=uniform_rv(a=1., b=5.))),
        (true_cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=double_exp_rv(pa=.3, a=1., b=5.))),
        (true_cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=scipy.stats.trapz(c=0.2, d=0.4))),
        (true_cpoisson_source,
         dict(vshape=(3, 2), lam=((1.,), (2.,)),
              y=scipy.stats.expon(scale=2.)))
        ]

    source_and_params_int = [
        (poisson_source, dict(vshape=())),
        (poisson_source, dict(vshape=(2,), lam=1.)),
        (poisson_source, dict(vshape=(2,), lam=((1.,), (2.,)))),
        (poisson_source, dict(vshape=(3, 2), lam=((1.,), (2.,)))),
        (poisson_source, dict(vshape=(2, 3, 5),
                              lam=np.arange(15.).reshape(3, 5, 1))),
        (true_poisson_source, dict(vshape=())),
        (true_poisson_source, dict(vshape=(2,), lam=1.)),
        (true_poisson_source, dict(vshape=(2,), lam=((1.,), (2.,)))),
        (true_poisson_source, dict(vshape=(3, 2), lam=((1.,), (2.,)))),
        (true_poisson_source, dict(vshape=(2, 3, 5),
                                   lam=np.arange(15.).reshape(3, 5, 1))),
        ]

    source_and_params_shortcuts = [
        (dw, dict(vshape=())),
        (dw, dict(vshape=(3, 2,), corr=((1., -.5), (-.5, 1.)))),
        (odd_dw, dict(vshape=())),
        (odd_dw, dict(vshape=(3, 2,), corr=((1., -.5), (-.5, 1.)))),
        (true_dw, dict(vshape=())),
        (true_dw, dict(vshape=(3, 2,), corr=((1., -.5), (-.5, 1.)))),
        (dj, dict(vshape=())),
        (dj, dict(vshape=(3, 2), lam=((1.,), (2.,)),
                  y=scipy.stats.expon(scale=2.))),
        (even_dj, dict(vshape=())),
        (even_dj, dict(vshape=(3, 2), lam=((1.,), (2.,)),
                       y=scipy.stats.expon(scale=2.))),
        (true_dj, dict(vshape=())),
        (true_dj, dict(vshape=(3, 2), lam=((1.,), (2.,)),
                       y=scipy.stats.expon(scale=2.))),
        ]

    source_and_params_shortcuts_int = [
        (dn, dict(vshape=())),
        (dn, dict(vshape=(2, 3, 5),
                  lam=np.arange(15.).reshape(3, 5, 1))),
        (even_dn, dict(vshape=())),
        (even_dn, dict(vshape=(2, 3, 5),
                       lam=np.arange(15.).reshape(3, 5, 1))),
        (true_dn, dict(vshape=())),
        (true_dn, dict(vshape=(2, 3, 5),
                       lam=np.arange(15.).reshape(3, 5, 1))),
        ]

    do(source_general, paths, float_dtype, source_and_params)
    do(source_general, paths, int_dtype, source_and_params_int)

    do(source_general, paths, float_dtype, source_and_params_shortcuts)
    do(source_general, paths, int_dtype, source_and_params_shortcuts_int)


# cases
def source_general(paths, dtype, source_and_params):
    cls, params = source_and_params

    src = cls(paths=paths, dtype=dtype, **params)
    assert_(src.paths == paths)
    assert_(src.vshape == params['vshape'])

    for dt in [2., (5.,), (0, 1.), np.linspace(0.5, 4., 12),
               np.arange(30.).reshape(2, 3, 5)]:
        s = src(0*np.asarray(dt), dt)
        assert_(s.shape == np.asarray(dt).shape +
                src.vshape + (src.paths,))
        assert_(s.dtype == dtype)

    if isinstance(src, true_source):
        for t in [2., (5.,), (0, 1.), np.linspace(0.5, 4., 12),
                  np.arange(30.).reshape(2, 3, 5)]:
            s = src(t)
            assert_(s.shape == np.asarray(t).shape +
                    src.vshape + (src.paths,))
            assert_(s.dtype == dtype)
            s = src(t, np.asarray(t)/10)
            sz = src.size
            assert_(sz > 1)
            st = src.t
            assert_(len(st) > 1)
    elif isinstance(src, source):
        assert_(src.size == 0)
        assert_(src.t.size == 0)
        assert_raises(TypeError, lambda: src(0.))
    else:
        assert_(False)


# ---------------------
# source specific tests
# ---------------------

# main test
def test_source_specific():
    np.random.seed(SEED)

    # wiener tests
    src = wiener_source(vshape=3, paths=5)
    assert_(src(0., 1.).shape == (3, 5))
    assert_raises(ValueError, wiener_source,
                  vshape=3, paths=10, corr=((1., .5), (.5, 1.)))

    # antithetics tests
    for src_class, sign in ((odd_wiener_source, -1),
                            (even_poisson_source, 1),
                            (even_cpoisson_source, 1)):
        src = src_class(vshape=(2, 3), paths=10)
        s = src(0., 1.)
        assert_array_equal(s[..., :5], sign*s[..., 5:])
        assert_raises(ValueError, src_class, paths=3)

    # poission tests
    assert_raises(ValueError, poisson_source,
                  vshape=(3, 2), lam=(1., 2., 3.))

    # compound poisson tests
    val = 2.

    class const_rv:
        def rvs(self, size):
            return np.full(size, fill_value=val)

    src = cpoisson_source(lam=1., paths=100, ptype=np.int16,
                          y=const_rv())
    s, n = src(0, 1), src.dn_value
    assert_allclose(n*val, s, rtol=eps(s.dtype))
    s, n = src(0, 100), src.dn_value
    assert_allclose(n*val, s, rtol=eps(s.dtype))

    # true sources tests
    for cls in (true_wiener_source, true_poisson_source, true_cpoisson_source,
                true_dw, true_dn, true_dj):
        src = cls(vshape=(2, 3), paths=10, t0=1., z0=2.)
        size1 = src.size
        assert_((src(1.) == 2.).all())
        s = src(2.)
        assert_(src.size >= 2*size1)
        assert_array_equal(src(2.), s)
        for dt in [-0.1, -1e-6, 0, 1e-6, 0.1]:
            z = src(2 + dt)
            w = src(2) + src(2, dt)
            assert_allclose(z, w, rtol=eps(z.dtype))
            if cls is true_wiener_source:
                nsigma = 5
                assert_((np.abs(z-s) <= sqrt(np.abs(dt))*nsigma).all())
        assert_(src[0, 0](2.).shape == (10,))
        assert_(src[:, :2](2., .1).shape == (2, 2, 10))
        assert_allclose(src[:1](2., .1), src(2., .1)[:1], rtol=eps(src.dtype))
        src2 = src[:, :, np.newaxis]
        assert_(src2.paths == src.paths)
        assert_(src2.dtype == src.dtype)
        assert_(src2.vshape == (2, 3, 1))
        assert_(src2(1).shape == (2, 3, 1, 10))

    # more true source tests
    for cls in (true_poisson_source, true_cpoisson_source,
                true_dn, true_dj):
        # check poisson sources do not store points between constant values
        src = cls(lam=0.)
        src(1.)
        size = src.size
        src(np.linspace(0, 1, 10))
        assert_(src.size == size)


# ------------------------------------------
# tests for time-dependent source parameters
# ------------------------------------------

# sources with time-dependent parameters are tested
# in test_processes, as sources of processes with
# time-dependent parameters

def corr(a, b=None):
    if b is None:
        a, b = a[0], a[1]
    cov = np.cov(a, b)
    n = cov.shape[0]
    return np.asarray(
        [[cov[i, j]/sqrt(cov[i, i]*cov[j, j]) for i in range(n)]
         for j in range(n)])


@slow
@quant
def test_source_true_wiener():

    # test exactness of true_wiener with correlations linarly
    # dependent on t
    rho0, irho0 = (.3, lambda t, t0: .3)
    rho1, irho1 = (lambda t: -0.1 - t/6, lambda t, t0: -0.1 - (t + t0)/12)
    rho2, irho2 = (lambda t: 0.1 + t/6, lambda t, t0: 0.1 + (t + t0)/12)
    rho3, irho3 = (None, lambda t, t0: 0)

    np.random.seed(SEED)
    for t0 in (0, 1):
        for rho, irho in ((rho0, irho0), (rho1, irho1),
                          (rho2, irho2), (rho3, irho3)):
            for vshape, i in zip(((2,), (3, 2)),
                                 (np.index_exp[...], np.index_exp[0])):
                tw = true_dw(rho=rho, vshape=vshape, paths=100000, t0=t0)
                for tt in (((6, 2, 4), (-6, -2, -4))):
                    tw(tt)
                    for s in tt:
                        x = tw(s)[i]
                        c = corr(x)
                        v = np.var(x)
                        assert_allclose(c[0, 1], irho(s, t0),
                                        rtol=0.02, atol=0.01)
                        assert_allclose(v, abs(s - t0), rtol=0.02, atol=0.01)
                        print('.', sep='', end='', file=sys.stderr)
                for s1, s2 in ((4, 6), (2, 6), (2, 4),
                               (-4, -6), (-2, -6), (-2, -4)):
                    c = corr((tw(s2) - tw(s1))[i][0], tw(s1)[i][0])
                    v = np.var((tw(s2) - tw(s1))[i][0])
                    assert_allclose(c[0, 1], 0, rtol=0.02, atol=0.01)
                    assert_allclose(v, abs(s2 - s1), rtol=0.02, atol=0.01)
                    print('.', sep='', end='', file=sys.stderr)



"""
=============================
COMMON TESTING INFRASTRUCTURE
=============================
"""
# common imports
import numpy as np
from numpy.testing import (
    assert_, assert_raises, assert_warns,
    assert_array_equal, assert_allclose,
    dec
    )
import itertools
import os
import sys
import warnings


# ----------------------------
# *** PACKAGE TO BE TESTED ***
# ----------------------------

# in all test modules, 'sp' is the package
# to be tested
import sdepy as sp
from sdepy import _config


# ---------------------------
# set up plotting if required
# ---------------------------

if _config.PLOT:
    import matplotlib.pyplot as plt

    # figure size
    plt.rcParams['figure.figsize'] = 12, 6


# ----------------------
# common test parameters
# ----------------------

KFUNC = _config.KFUNC

# exact equality is tested up to the float resolution times EPS_FACTOR
# (see eps function below)
EPS_FACTOR = 16

# each testing routine should call np.random.seed(SEED)
SEED = 1234

# DIR is the directory used, in quantitative tests,
# to retrieve expected errors and to save realized errors and plots
PACKAGE_DIR = os.path.split(sp.__file__)[0]
DIR = os.path.join(PACKAGE_DIR, 'tests', 'cfr')

# Saving flags
# PLOT = False disables plots construction and saving
# (each testing module checks this flag and behaves accordingly).
# SAVE_ERRORS = False disables saving realized errors
# (see save_errors() below).
PLOT = _config.PLOT
SAVE_ERRORS = _config.SAVE_ERRORS

# testing flags
QUANT_TEST_FAIL = _config.QUANT_TEST_FAIL
VERBOSE = _config.VERBOSE

# Relative tolerance when checking expected vs. realized errors:
# the expected errors to be tested, returned by load_errors,
# are the stored expected error times (1 + EXPECTED_ERROR_RTOL).
EXPECTED_ERROR_RTOL = 0.01

# High or low definition of quantitative tests
QUANT_TEST_MODE = _config.QUANT_TEST_MODE
assert QUANT_TEST_MODE in {'HD', 'LD'}
PATHS_HD = 100*1000
PATHS_LD = 100


# --------------------------
# functions for common tasks
# --------------------------

def quant(test):
    """Decorator to mark test as quantitative,
    setting the attribute test.quant = True
    """
    test.quant = True
    return test

slow = dec.slow


def do(testing_function, *case_iterators, **args):
    """call a test function f across all test cases exposed
    in case_iterators, passing keyword arguments in args"""

    for z in case_iterators:
        # enforce predictability of test order
        assert isinstance(z, (list, tuple))

    for case in itertools.product(*case_iterators):
        testing_function(*case, **args)
        print('.', sep='', end='', file=sys.stderr)


def eps(dtype):
    if np.dtype(dtype).kind in ('f', 'c'):
        return np.finfo(dtype).resolution * EPS_FACTOR
    else:
        return 0


def choice(list_):
    return list_[np.random.randint(len(list_))]


class const_errors:

    def __init__(self, error):
        self.error = error

    def __getitem__(self, key):
        return (self.error, self.error)

    def __setitem__(self, key, value):
        pass

noerrors_expected = const_errors(1000)
noerrors_realized = const_errors(0)


def load_errors(context, dtype=float):
    fname = os.path.join(DIR, context + '_err_expected.txt')
    min_error = eps(float)

    def err(error):
        return max(min_error, abs(float(error)*(1 + EXPECTED_ERROR_RTOL)))

    if QUANT_TEST_FAIL:
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                lines = [line.split() for line in f.readlines()]
                errors = {test_key: (err(mean_error), err(max_error))
                          for test_key, mean_error, max_error in lines[2:]}
        else:
            warnings.warn('No expected errors file found while testing {}: '
                          'file {} does not exist'.format(context, fname),
                          RuntimeWarning)
            errors = noerrors_expected
    else:
        errors = noerrors_expected

    return errors


def save_errors(context, errors):
    if SAVE_ERRORS:
        fname = os.path.join(DIR, context + '_err_realized.txt')
        with open(fname, 'w') as f:
            print('{:40} {:>12} {:>12}\n'.format(
                'TEST_ID', 'MEAN_ERROR', 'MAX_ERROR'), file=f)
            for test_key, (mean_error, max_error) in errors.items():
                print('{:40s} {:12.5g} {:12.5g}'.format(
                    test_key, mean_error, max_error), file=f)


def save_figure(fig, fig_id):
    fname = os.path.join(DIR, fig_id + '.png')
    fig.savefig(fname, dpi=300)


def plot_histogram(hist, **kwargs):
    counts, bins = hist
    return plt.hist(bins[:-1], bins=bins, weights=counts, **kwargs)


# -----------------------
# assert_ for quant tests
# -----------------------

message = """
A quantitative test has failed, running in {} definition with {} paths.
This test relies on the reproducibility of expected errors, once
random numbers have been seeded with np.random.seed({}), and its failure
may not necessarily indicate that the package is broken.
Consider changing the test configuration in
{} by setting

    QUANT_TEST_FAIL = False
    QUANT_TEST_MODE = 'HD'
    PLOT = True

and inspecting the plots that are generated in the folder
{}."""


def assert_quant(flag):

    if not flag:
        definition, paths = (
            ('low', PATHS_LD) if _config.QUANT_TEST_MODE == 'LD'
            else ('high', PATHS_HD))

        warnings.warn(
            message.format(definition, paths, SEED,
                           os.path.join(PACKAGE_DIR, '_config.py'),
                           DIR),
            RuntimeWarning)

    assert_(flag)



"""
==============================
PRIVATE CONFIGURATION SETTINGS
==============================

Global settings for:
*  ``kfunc`` deployment,
*  unit testing behaviour.
"""

# kfunc deployment (see .shortcuts)
KFUNC = 'shortcuts'

# testing (see .tests.shared)
PLOT = False
SAVE_ERRORS = False
VERBOSE = False
QUANT_TEST_MODE = 'LD'
QUANT_TEST_FAIL = True



"""
=========================================
ANALYTICAL RESULTS FOR REALIZED PROCESSES
=========================================
"""
import numpy as np
import scipy
from numpy import sqrt, exp, log


###################################
# Statistics for realized processes
###################################

# when used for multi-dimensional processes,
# these analytical results are valid only if there is
# no correlation among the process components.

# --------------------
# wiener_process stats
# --------------------

# @kfunc(nvar=1)
def wiener_mean(t, *, x0=0., mu=0., sigma=1.):
    """
    wiener_mean(t, *, x0=0., mu=0., sigma=1.)

    Mean of values at time t of a Wiener process
    (as per the wiener_process class) with
    time-independent parameters.

    See Also
    --------
    wiener_process
    """
    t, x0, mu, sigma = np.broadcast_arrays(t, x0, mu, sigma)
    return x0 + mu*t


# @kfunc(nvar=1)
def wiener_var(t, *, x0=0., mu=0., sigma=1.):
    """
    wiener_var(t, *, x0=0., mu=0., sigma=1.)

    Variance of values at time t of a Wiener process
    (as per the wiener_process class) with
    time-independent parameters.

    See Also
    --------
    wiener_process
    """
    t, x0, mu, sigma = np.broadcast_arrays(t, x0, mu, sigma)
    return sigma*sigma*t


# @kfunc(nvar=1)
def wiener_std(t, *, x0=0., mu=0., sigma=1.):
    """
    wiener_std(t, *, x0=0., mu=0., sigma=1.)

    Standard deviation of values at time t of a Wiener process
    (as per the wiener_process class) with
    time-independent parameters.

    See Also
    --------
    wiener_process
    """
    return sqrt(wiener_var(t, x0=x0, mu=mu, sigma=sigma))


# @kfunc(nvar=2)
def wiener_pdf(t, x, *, x0=0., mu=0., sigma=1.):
    """
    wiener_pdf(t, x, *, x0=0., mu=0., sigma=1.)

    Probability distribution function of values
    at time t of a Wiener process (as per the wiener_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    wiener_process
    """
    t, x, x0, mu, sigma = np.broadcast_arrays(t, x, x0, mu, sigma)
    mean = x0 + mu*t
    std = sigma*sqrt(t)
    return scipy.stats.norm.pdf(x, loc=mean, scale=std)


# @kfunc(nvar=2)
def wiener_cdf(t, x, *, x0=0., mu=0., sigma=1.):
    """
    wiener_cdf(t, x, *, x0=0., mu=0., sigma=1.)

    Cumulative probability distribution function of values
    at time t of a Wiener process (as per the wiener_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    wiener_process
    """
    t, x, x0, mu, sigma = np.broadcast_arrays(t, x, x0, mu, sigma)
    mean = x0 + mu*t
    std = sigma*sqrt(t)
    return scipy.stats.norm.cdf(x, loc=mean, scale=std)


# @kfunc(nvar=2)
def wiener_chf(t, u, *, x0=0., mu=0., sigma=1.):
    """
    wiener_chf(t, u, *, x0=0., mu=0., sigma=1.)

    Characteristic function of the probability distribution of values
    at time t of a Wiener process (as per the wiener_process class)
    with time-independent parameters, evaluated at u.

    See Also
    --------
    wiener_process
    """
    t, u, x0, mu, sigma = np.broadcast_arrays(t, u, x0, mu, sigma)
    return exp(1j*u*x0 + t*(1j*mu*u - sigma*sigma*u*u/2))


# ---------------------
# lognorm_process stats
# ---------------------

# @kfunc(nvar=1)
def lognorm_mean(t, *, x0=1., mu=0., sigma=1.):
    """
    lognorm_mean(t, *, x0=1., mu=0., sigma=1.)

    Mean of values at time t of a lognormal process
    (as per the lognorm_process class) with time-independent parameters.

    See Also
    --------
    lognorm_process
    """
    t, x0, mu, sigma = np.broadcast_arrays(t, x0, mu, sigma)
    return x0 * exp(mu*t)


# @kfunc(nvar=1)
def lognorm_var(t, *, x0=1., mu=0., sigma=1.):
    """
    lognorm_var(t, *, x0=1., mu=0., sigma=1.)

    Variance of values at time t of a lognormal process
    (as per the lognorm_process class) with time-independent parameters.

    See Also
    --------
    lognorm_process
    """
    t, x0, mu, sigma = np.broadcast_arrays(t, x0, mu, sigma)
    return x0*x0 * (exp(sigma*sigma*t) - 1) * exp(2*mu*t)


# @kfunc(nvar=1)
def lognorm_std(t, *, x0=1., mu=0., sigma=1.):
    """
    lognorm_std(t, *, x0=1., mu=0., sigma=1.)

    Standard deviation of values at time t of a lognormal process
    (as per the lognorm_process class) with time-independent parameters.

    See Also
    --------
    lognorm_process
    """
    return sqrt(lognorm_var(t, x0=x0, mu=mu, sigma=sigma))


# @kfunc(nvar=2)
def lognorm_pdf(t, x, *, x0=1., mu=0., sigma=1.):
    """
    lognorm_pdf(t, x, *, x0=1., mu=0., sigma=1.)

    Probability distribution function of values at time t of a
    lognormal process (as per the lognorm_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    lognorm_process
    """
    t, x, x0, mu, sigma = np.broadcast_arrays(t, x, x0, mu, sigma)
    mt, st = mu*t, sigma*sqrt(t)
    return scipy.stats.lognorm.pdf(x, s=st,
                                   scale=x0*exp(mt - st*st/2))


# @kfunc(nvar=2)
def lognorm_cdf(t, x, *, x0=1., mu=0., sigma=1.):
    """
    lognorm_cdf(t, x, *, x0=1., mu=0., sigma=1.)

    Cumulative probability distribution function of values
    at time t of a lognormal process (as per the lognorm_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    lognorm_process
    """
    t, x, x0, mu, sigma = np.broadcast_arrays(t, x, x0, mu, sigma)
    mt, st = mu*t, sigma*sqrt(t)
    return scipy.stats.lognorm.cdf(x, s=st,
                                   scale=x0*exp(mt - st*st/2))


# @kfunc(nvar=2)
def lognorm_log_chf(t, u, *, x0=1., mu=0., sigma=1.):
    """
    lognorm_log_chf(t, u, *, x0=1., mu=0., sigma=1.)

    Characteristic function of the probability distribution of values
    at time t of the logarithm of a lognormal process (as per the
    lognorm_process class) with time-independent parameters, evaluated at u.

    See Also
    --------
    lognorm_process
    """
    t, u, x0, mu, sigma = np.broadcast_arrays(t, u, x0, mu, sigma)
    mu0 = mu - sigma*sigma/2
    return exp(1j*u*log(x0) + t*(1j*mu0*u - sigma*sigma*u*u/2))


# --------------------------------
# ornstein_uhlenbeck_process stats
# --------------------------------

# @kfunc(nvar=1)
def oruh_mean(t, *, x0=0., theta=0., k=1., sigma=1.):
    """
    oruh_mean(t, *, x0=0., theta=0., k=1., sigma=1.)

    Mean of values at time t of an Ornstein-Uhlenbeck process
    (as per the ornstein_uhlenbeck_process class) with
    time-independent parameters.

    See Also
    --------
    ornstein_uhlenbeck_process
    """
    t, x0, theta, k, sigma = np.broadcast_arrays(t, x0, theta, k, sigma)
    return theta - exp(-k*t) * (theta - x0)


# @kfunc(nvar=1)
def oruh_var(t, *, x0=0., theta=0., k=1., sigma=1.):
    """
    oruh_var(t, *, x0=0., theta=0., k=1., sigma=1.)

    Variance of values at time t of an Ornstein-Uhlenbeck process
    (as per the ornstein_uhlenbeck_process class) with
    time-independent parameters.

    See Also
    --------
    ornstein_uhlenbeck_process
    """
    t, x0, theta, k, sigma = np.broadcast_arrays(t, x0, theta, k, sigma)
    return (sigma*sigma)/(2*k)*(1 - exp(-2*k*t))


# @kfunc(nvar=1)
def oruh_std(t, *, x0=0., theta=0., k=1., sigma=1.):
    """
    oruh_std(t, *, x0=0., theta=0., k=1., sigma=1.)

    Standard deviation of values at time t of an Ornstein-Uhlenbeck
    process (as per the ornstein_uhlenbeck_process class)
    with time-independent parameters.

    See Also
    --------
    ornstein_uhlenbeck_process
    """
    return sqrt(oruh_var(t, x0=x0, theta=theta, k=k, sigma=sigma))


# @kfunc(nvar=2)
def oruh_pdf(t, x, *, x0=0., theta=0., k=1., sigma=1.):
    """
    oruh_pdf(t, x, *, x0=0., theta=0., k=1., sigma=1.)

    Probability distribution function of values at time t of an
    Ornstein-Uhlenbeck process (as per the ornstein_uhlenbeck_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    ornstein_uhlenbeck_process
    """
    mean = oruh_mean(t, x0=x0, theta=theta, k=k, sigma=sigma)
    var = oruh_var(t, x0=x0, theta=theta, k=k, sigma=sigma)
    return scipy.stats.norm.pdf(x, loc=mean, scale=sqrt(var))


# @kfunc(nvar=2)
def oruh_cdf(t, x, *, x0=0., theta=0., k=1., sigma=1.):
    """
    oruh_cdf(t, x, *, x0=0., theta=0., k=1., sigma=1.)

    Cumulative probability distribution function of values at time t of an
    Ornstein-Uhlenbeck process (as per the ornstein_uhlenbeck_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    ornstein_uhlenbeck_process
    """
    mean = oruh_mean(t, x0=x0, theta=theta, k=k, sigma=sigma)
    var = oruh_var(t, x0=x0, theta=theta, k=k, sigma=sigma)
    return scipy.stats.norm.cdf(x, loc=mean, scale=sqrt(var))


# -----------------------------------
# hull_white_process stats, factors=1
# -----------------------------------

hw1f_mean = oruh_mean
hw1f_var = oruh_var
hw1f_std = oruh_std
hw1f_pdf = oruh_pdf
hw1f_cdf = oruh_cdf


# -----------------------------------
# hull_white_process stats, factors=2
# -----------------------------------

def _hw2f_args_setup(x0, theta, k, sigma, *var):

    x0, theta, k, sigma = np.broadcast_arrays(x0, theta, k, sigma)
    shape = x0.shape
    if len(shape) == 0:
        check = True
        x1, x2 = x0, x0
        theta1, theta2 = theta, theta
        k1, k2 = k, k
        s1, s2 = sigma, sigma
    elif len(shape) == 1:
        check = (shape[-1] in (1, 2))
        x1, x2 = x0[0], x0[-1]
        theta1, theta2 = theta[0], theta[-1]
        k1, k2 = k[0], k[-1]
        s1, s2 = sigma[0], sigma[-1]
    else:
        check = (shape[-2] in (1, 2))
        x1, x2 = x0[..., 0, :], x0[..., -1, :]
        theta1, theta2 = theta[..., 0, :], theta[..., -1, :]
        k1, k2 = k[..., 0, :], k[..., -1, :]
        s1, s2 = sigma[..., 0, :], sigma[..., -1, :]

    if not check:
        raise ValueError(
            'for a 2 factor Hull-White process, '
            'each of x0, theta, k, sigma should be scalar or '
            'a 2-component vector (representing factors) or '
            'have axis -2 of size 1 or 2 (representing factors) '
            'and axis -1 of size 1 (matching the paths axis of '
            'hull_white_process output)'
        )

    return np.broadcast_arrays(x1, x2, theta1, theta2, k1, k2, s1, s2, *var)


# @kfunc(nvar=1)
def hw2f_mean(t, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
              sigma=(1., 1.), rho=0.):
    """
    hw2f_mean(t, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
              sigma=(1., 1.), rho=0.)

    Mean of values at time t of a Hull-White 2-factors process
    (as per the hull_white_process class) with time-independent parameters.

    See Also
    --------
    hull_white_process
    """
    x1, x2, theta1, theta2, k1, k2, s1, s2, rho, t = \
        _hw2f_args_setup(x0, theta, k, sigma, rho, t)

    return (theta1 - exp(-k1*t) * (theta1 - x1) +
            theta2 - exp(-k2*t) * (theta2 - x2))


# @kfunc(nvar=1)
def hw2f_var(t, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.):
    """
    hw2f_var(t, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.)

    Variance of values at time t of a Hull-White 2-factors process
    (as per the hull_white_process class) with time-independent parameters.

    See Also
    --------
    hull_white_process
    """
    x1, x2, theta1, theta2, k1, k2, s1, s2, rho, t = \
        _hw2f_args_setup(x0, theta, k, sigma, rho, t)

    return (s1*s1)/(2*k1)*(1 - exp(-2*k1*t)) + \
           (s2*s2)/(2*k2)*(1 - exp(-2*k2*t)) + \
           (2*s1*s2*rho)/(k1 + k2)*(1 - exp(-(k1 + k2)*t))


# @kfunc(nvar=1)
def hw2f_std(t, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.):
    """
    hw2f_std(t, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.)

    Standard deviation of values at time t of a Hull-White 2-factors
    process (as per the hull_white_process class)
    with time-independent parameters.

    See Also
    --------
    hull_white_process
    """
    return sqrt(hw2f_var(t, x0=x0, theta=theta, k=k,
                         sigma=sigma, rho=rho))


# @kfunc(nvar=2)
def hw2f_pdf(t, x, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.):
    """
    hw2f_pdf(t, x, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.)

    Probability distribution function of values at time t of a
    Hull-White 2-factors process (as per the hull_white_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    hull_white_process
    """
    x1, x2, theta1, theta2, k1, k2, s1, s2, rho, t, x = \
        _hw2f_args_setup(x0, theta, k, sigma, rho, t, x)
    mean = hw2f_mean(t, x0=x0, theta=theta, k=k, sigma=sigma, rho=rho)
    var = hw2f_var(t, x0=x0, theta=theta, k=k, sigma=sigma, rho=rho)
    return scipy.stats.norm.pdf(x, loc=mean, scale=sqrt(var))


# @kfunc(nvar=2)
def hw2f_cdf(t, x, *, x0=(0., 0.), theta=(0., 0.), k=(1., 1.),
             sigma=(1., 1.), rho=0.):
    """Cumulative probability distribution function of values at time t
    of a Hull-White 2-factors process (as per the hull_white_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    hull_white_process
    """
    x1, x2, theta1, theta2, k1, k2, s1, s2, rho, t, x = \
        _hw2f_args_setup(x0, theta, k, sigma, rho, t, x)
    mean = hw2f_mean(t, x0=x0, theta=theta, k=k, sigma=sigma, rho=rho)
    var = hw2f_var(t, x0=x0, theta=theta, k=k, sigma=sigma, rho=rho)
    return scipy.stats.norm.cdf(x, loc=mean, scale=sqrt(var))


# --------------------------------
# cox_ingersoll_ross_process stats
# --------------------------------

# @kfunc(nvar=1)
def cir_mean(t, *, x0=1., theta=1., k=1., xi=1.):
    """
    cir_mean(t, *, x0=1., theta=1., k=1., xi=1.)

    Mean of values at time t of a Cox-Ingersoll-Ross process (as per the
    cox_ingersoll_ross_process class) with time-independent parameters.

    See Also
    --------
    cox_ingersoll_ross_process
    """
    t, x0, theta, k, xi = np.broadcast_arrays(t, x0, theta, k, xi)
    alpha = exp(-k*t)
    return theta + (x0 - theta) * alpha


# @kfunc(nvar=1)
def cir_var(t, *, x0=1., theta=1., k=1., xi=1.):
    """
    cir_var(t, *, x0=1., theta=1., k=1., xi=1.)

    Variance of values at time t of a Cox-Ingersoll-Ross process (as per the
    cox_ingersoll_ross_process class) with time-independent parameters.

    See Also
    --------
    cox_ingersoll_ross_process
    """
    t, x0, theta, k, xi = np.broadcast_arrays(t, x0, theta, k, xi)
    alpha = exp(-k*t)
    return (xi*xi / k) * (1 - alpha) * \
        (theta/2 + (x0 - theta/2) * alpha)


# @kfunc(nvar=1)
def cir_std(t, *, x0=1., theta=1., k=1., xi=1.):
    """
    cir_std(t, *, x0=1., theta=1., k=1., xi=1.)

    Standard deviation of values at time t of a Cox-Ingersoll-Ross process
    (as per the cox_ingersoll_ross_process class)
    with time-independent parameters.

    See Also
    --------
    cox_ingersoll_ross_process
    """
    return sqrt(cir_var(t, x0=x0, theta=theta, k=k, xi=xi))


# @kfunc(nvar=2)
def cir_pdf(t, x, *, x0=1., theta=1., k=1., xi=1.):
    """
    cir_pdf(t, x, *, x0=1., theta=1., k=1., xi=1.)

    Probability distribution function of values at time t of a
    Cox-Ingersoll-Ross process (as per the cox_ingersoll_ross_process class)
    with time-independent parameters, evaluated at x.

    See Also
    --------
    cox_ingersoll_ross_process
    """
    t, x0, theta, k, xi = np.broadcast_arrays(t, x0, theta, k, xi)
    alpha = exp(-k*t)
    c = 2*k / (1 - alpha) / (xi*xi)
    q = 2*k * theta / (xi*xi) - 1
    return c*exp(-c * (x0*alpha + x)) * \
        (x / (x0*alpha))**(q/2) * \
        scipy.special.iv(q, 2*c * sqrt(x*x0*alpha))


# --------------------
# heston_process stats
# --------------------

# @kfunc(nvar=1)
def heston_log_mean(t, *, x0=1., mu=0., sigma=1.,
                    y0=1., theta=1., k=1.,
                    xi=1., rho=0.):
    """
    heston_log_mean(t, *, x0=1., mu=0., sigma=1., y0=1., theta=1., k=1.,
                    xi=1., rho=0.)

    Mean of the logarithm of values at time t of a Heston process
    (as per the full_heston_process class) with time-independent parameters.

    See Also
    --------
    full_heston_process
    """
    t, x0, mu, sigma, y0, theta, k, xi, rho = \
        np.broadcast_arrays(t, x0, mu, sigma, y0, theta, k, xi, rho)
    alpha = exp(-k*t)
    return log(x0) + \
        sigma*sigma * (theta - y0) * (1-alpha) / (2*k) + \
        (mu - theta*sigma*sigma/2) * t


# @kfunc(nvar=1)
def heston_log_var(t, *, x0=1., mu=0., sigma=1.,
                   y0=1., theta=1., k=1.,
                   xi=1., rho=0.):
    """Variance of the logarithm of values at time t of a Heston process
    (as per the full_heston_process class) with time-independent parameters.

    See Also
    --------
    full_heston_process
    """
    t, x0, mu, sigma, y0, theta, k, xi, rho = \
        np.broadcast_arrays(t, x0, mu, sigma, y0, theta, k, xi, rho)
    alpha = exp(-k*t)
    omega1 = (alpha*sigma*xi)**2 + \
        4*alpha * ((1 + k*t)*(sigma*xi)**2 -
                   2*rho*k*sigma*xi*(2 + k*t) +
                   2*k*k) + \
        (2*k*t - 5) * (sigma*xi)**2 - \
        8*rho*k*sigma*xi * (k*t - 2) + \
        8*k*k * (k*t - 1)
    omega2 = -(alpha*sigma*xi)**2 + \
        2*alpha * (-(sigma*xi)**2 * k*t +
                   2*rho*sigma*xi*k * (k*t + 1) -
                   2*k*k) + \
        (sigma*xi)**2 - \
        4*rho*k*sigma*xi + \
        4*k*k

    return sigma*sigma*(theta*omega1/2 + y0*omega2) / (4*k**3)


# @kfunc(nvar=1)
def heston_log_std(t, *, x0=1., mu=0., sigma=1.,
                   y0=1., theta=1., k=1.,
                   xi=1., rho=0.):
    """
    heston_log_std(t, *, x0=1., mu=0., sigma=1., y0=1., theta=1., k=1.,
                   xi=1., rho=0.)

    Standard deviation of the logarithm of values at time t of a Heston
    process (as per the full_heston_process class)
    with time-independent parameters.

    See Also
    --------
    full_heston_process
    """
    return sqrt(heston_log_var(t, x0=x0, mu=mu, sigma=sigma,
                               y0=y0, theta=theta, k=k,
                               xi=xi, rho=rho))


# @kfunc(nvar=2)
def heston_log_chf(t, u, *, x0=1., mu=0., sigma=1.,
                   y0=1., theta=1., k=1.,
                   xi=1., rho=0.):
    """
    heston_log_chf(t, u, *, x0=1., mu=0., sigma=1., y0=1., theta=1., k=1.,
                   xi=1., rho=0.)

    Characteristic function of the probability distribution of values at
    time t of the logarithm of a Heston process (as per the full_heston_process
    class) , with time-independent parameters, evaluated at u.

    See Also
    --------
    full_heston_process
    """
    t, u, x0, mu, sigma, y0, theta, k, xi, rho = \
        np.broadcast_arrays(t, u, x0, mu, sigma, y0, theta, k, xi, rho)
    d = sqrt((rho*sigma*xi*u*1j - k)**2 + (sigma*xi)**2 * u*(1j + u))
    ap, am = (k - rho*sigma*xi*u*1.j - d), (k - rho*sigma*xi*u*1j + d)
    g = ap/am
    alpha = exp(-d*t)

    A = 1j*u*(log(x0) + mu*t)
    A += theta*k/(xi*xi) * (ap*t - 2*log((1 - g*alpha)/(1-g)))
    A += y0/(xi*xi) * ap * (1 - alpha)/(1 - g*alpha)
    return exp(A)


def _chf_to_pdf(t, x, chf, **chf_args):
    """
    Estimate by numerical integration, using ``scipy.integrate.quad``,
    of the probability distribution described by the given characteristic
    function. Integration errors are not reported/checked.
    Either ``t`` or ``x`` must be a scalar.
    """
    t = np.asarray(t)
    x = np.asarray(x)

    def f(u, t, x):
        return np.real(
            exp(-1j*u*x) / (2*np.pi) * chf(t, u, **chf_args))
    if t.shape != ():
        pdf = np.empty(t.shape)
        for i in np.ndindex(t.shape):
            pdf[i] = scipy.integrate.quad(
                lambda u: f(u, t[i], x), -np.inf, np.inf)[0]
    else:
        pdf = np.empty(x.shape)
        for i in np.ndindex(x.shape):
            pdf[i] = scipy.integrate.quad(
                lambda u: f(u, t, x[i]), -np.inf, np.inf)[0]
    return pdf


# @kfunc(nvar=2)
def heston_log_pdf(t, logx, *, x0=1., mu=0., sigma=1.,
                   y0=1., theta=1., k=1.,
                   xi=1., rho=0.):
    """
    heston_log_pdf(t, logx, *, x0=1., mu=0., sigma=1., y0=1., theta=1., k=1.,
                   xi=1., rho=0.)

    Probability distribution function of values at time t of the
    logarithm of a Heston process, (as per the full_heston_process class)
    with time-independent parameters, evaluated at logx.

    Notes
    -----
    Estimate by numerical integration, using ``scipy.integrate.quad``,
    of the closed-form characteristic function ``heston_log_chf``.
    Integration errors are not reported/checked. Either ``t`` or ``logx``
    must be a scalar.

    See Also
    --------
    full_heston_process
    """
    return _chf_to_pdf(t, logx, heston_log_chf,
                       x0=x0, mu=mu, sigma=sigma,
                       y0=y0, theta=theta,
                       k=k, xi=xi, rho=rho)


# -----------------------------------
# merton_jumpdiff_process stats
# -----------------------------------

# @kfunc(nvar=1)
def mjd_mean(t, *, x0=1., mu=0., sigma=1., lam=1., a=0.0, b=1.):
    """
    mjd_mean(t, *, x0=1., mu=0., sigma=1., lam=1., a=0.0, b=1.)

    Mean of values at time t of a Merton jump-diffusion process, (as per the
    merton_jumpdiff_process class) with time-independent parameters.

    See Also
    --------
    jumpdiff_process
    merton_jumpdiff_process
    """
    t, x0, mu, sigma, lam, a, b = \
        np.broadcast_arrays(t, x0, mu, sigma, lam, a, b)

    # martingale correction *NOT* applied
    exp_mean = exp(a + b*b/2)
    nu = lam*(exp_mean - 1)
    mu1 = mu + nu

    return x0*exp(mu1*t)


# @kfunc(nvar=2)
def mjd_log_chf(t, u, *, x0=1., mu=0., sigma=1., lam=1., a=0.0, b=1.):
    """
    mjd_log_chf(t, u, *, x0=1., mu=0., sigma=1., lam=1., a=0.0, b=1.)

    Characteristic function of the probability distribution of values
    at time t of the logarithm of a Merton jump-diffusion process
    (as per the merton_jumpdiff_process class), with
    time-independent parameters, evaluated at u.

    See Also
    --------
    jumpdiff_process
    merton_jumpdiff_process
    """

    t, u, x0, mu, sigma, lam, a, b = \
        np.broadcast_arrays(t, u, x0, mu, sigma, lam, a, b)

    # martingale correction *NOT* applied
    #   exp_mean = exp(a + b*b/2)
    #   nu = lam*(exp_mean - 1)
    #   mu0 = mu - sigma*sigma/2 - nu

    mu0 = mu - sigma*sigma/2
    A = 1j*u*log(x0) + \
        t*(+ 1j*mu0*u -
           sigma*sigma*u*u/2 +
           lam * (exp(1j*a*u - b*b*u*u/2) - 1))

    return exp(A)


# @kfunc(nvar=2)
def mjd_log_pdf(t, logx, *, x0=1., mu=0., sigma=1., lam=1., a=0.0, b=1.):
    """
    mjd_log_pdf(t, logx, *, x0=1., mu=0., sigma=1., lam=1., a=0.0, b=1.)

    Probability distribution function of values at time t of the logarithm
    of a Merton jump-diffusion process (as per the
    merton_jumpdiff_process class), with time-independent parameters,
    evaluated at logx.

    Notes
    -----
    Estimate by numerical integration, using ``scipy.integrate.quad``,
    of the closed-form characteristic function ``mjd_log_chf``.
    Integration errors are not reported/checked. Either ``t`` or ``logx``
    must be a scalar.

    See Also
    --------
    jumpdiff_process
    merton_jumpdiff_process
    mjd_log_chf
    """
    return _chf_to_pdf(t, logx, mjd_log_chf,
                       x0=x0, mu=mu, sigma=sigma,
                       lam=lam, a=a, b=b)


# --------------------------------
# kou_jumpdiff_process stats
# --------------------------------

# @kfunc(nvar=1)
def kou_mean(t, *, x0=1., mu=0., sigma=1., lam=1., a=0.5, b=0.5, pa=0.5):
    """
    kou_mean(t, *, x0=1., mu=0., sigma=1., lam=1., a=0.5, b=0.5, pa=0.5)

    Mean of values at time t of a double exponential (Kou)
    jump-diffusion process (as per the kou_jumpdiff_process class)
    with time-independent parameters.

    See Also
    --------
    jumpdiff_process
    kou_jumpdiff_process
    """
    t, x0, mu, sigma, lam, a, b, pa = \
        np.broadcast_arrays(t, x0, mu, sigma, lam, a, b, pa)
    pb = 1 - pa

    # martingale correction *NOT* applied
    exp_mean = pa/(1 - a) + pb/(1 + b)
    nu = lam*(exp_mean - 1)
    mu1 = mu + nu

    return x0*exp(mu1*t)


# @kfunc(nvar=2)
def kou_log_chf(t, u, *, x0=1., mu=0., sigma=1., lam=1., a=0.5, b=0.5, pa=0.5):
    """
    kou_log_chf(t, u, *, x0=1., mu=0., sigma=1., lam=1., a=0.5, b=0.5, pa=0.5)

    Characteristic function of the probability distribution of values
    at time t of the logarithm of a Kou jump-diffusion process,
    (as per the kou_jumpdiff_process class) with
    time-independent parameters, evaluated at u.

    See Also
    --------
    jumpdiff_process
    kou_jumpdiff_process
    """
    t, u, x0, mu, sigma, lam, a, b, pa = \
        np.broadcast_arrays(t, u, x0, mu, sigma, lam, a, b, pa)
    pb = 1 - pa

    # martingale correction *NOT* applied
    #   exp_mean = pa/(1 - a) + pb/(1 + b)
    #   nu = lam*(exp_mean - 1)
    #   mu0 = mu - sigma*sigma/2 - nu

    mu0 = mu - sigma*sigma/2
    A = 1j*u*log(x0) + \
        t*(+ 1j*mu0*u -
           sigma*sigma*u*u/2 +
           1j*u*lam*(pa/(1/a - 1j*u) - pb/(1/b + 1j*u)))

    return exp(A)


# @kfunc(nvar=2)
def kou_log_pdf(t, logx, *, x0=1., mu=0., sigma=1., lam=1.,
                pa=0.5, a=0.5, b=0.5):
    """
    kou_log_pdf(t, logx, *, x0=1., mu=0., sigma=1., lam=1.,
                pa=0.5, a=0.5, b=0.5)

    Probability distribution function of values at time t of the logarithm
    of a double-exponential (Kou) jump-diffusion process (as per the
    kou_jumpdiff_process class), with time-independent parameters,
    evaluated at logx.

    Notes
    -----
    Estimate by numerical integration, using ``scipy.integrate.quad``,
    of the closed-form characteristic function ``kou_log_chf``.
    Integration errors are not reported/checked. Either ``t`` or ``logx``
    must be a scalar.

    See Also
    --------
    jumpdiff_process
    kou_jumpdiff_process
    kou_log_chf
    """
    return _chf_to_pdf(t, logx, kou_log_chf,
                       x0=x0, mu=mu, sigma=sigma,
                       lam=lam, a=a, b=b, pa=pa)


#################################################
# Black-Scholes formulae for call and put options
#################################################

# @kfunc(nvar=2)
def bsd1d2(k, t, *, x0=1., r=0., q=0., sigma=1.):
    """
    bsd1d2(k, t, *, x0=1., r=0., q=0., sigma=1.)

    Black-Scholes d1 and d2 coefficients.

    See Also
    --------
    bscall
    """
    # convert inputs to arrays
    k, t, x0, r, q, sigma = np.broadcast_arrays(
        k, t, x0, r, q, sigma)

    # compute d1 and d2, handling limiting cases t==0 and/or x0==k
    # and/or sigma==0
    A = log(x0/k) + (r - q + 0.5 * sigma**2) * t
    B = sigma * sqrt(t)
    old_settings = np.seterr(divide='ignore', invalid='ignore')
    d1 = np.where(B == 0, np.where(A == 0, 0., np.inf * A), A / B)
    np.seterr(**old_settings)
    assert not np.isnan(d1).any()
    d2 = d1 - B

    return d1, d2


# @kfunc(nvar=2)
def bscall(k, t, *, x0=1., r=0., q=0., sigma=1.):
    """
    bscall(k, t, *, x0=1., r=0., q=0., sigma=1.)

    Black-Scholes call option value.

    Parameters
    ----------
    k : array-like
        Strike.
    t : array-like
        Time to maturity.
    x0 : array-like
        Initial value of underlying security.
    r : array-like
        Risk-free rate.
    q : array-like
        Dividend yield of underlying security.
    sigma : array-like
        Volatility of underlying security.

    Returns
    -------
    array
        Risk neutral valuation at time s=0 of an European call option
        paying ``max(x(t) - k, 0)`` at maturity, where the price ``x(s)``
        of the underlying security follows a lognormal process
        with ``x(0) = x0`` and volatility ``sigma``.

    See Also
    --------
    bsd1d2
    bscall_delta
    bsput
    bsput_delta

    Notes
    -----
    ``bscall(k, t, x0, r, q, sigma)`` returns::

        bscall_value = x0*exp(-q*t)*norm.cdf(d1) + k*exp(-r*t)*norm.cdf(d2)

    where ``cdf`` is ``scipy.stats.norm.cdf`` and
    ``d1, d2 = bsd1d2(k, t, x0, r, q, sigma)`` are given as::

        d1 = (log(x0/k) + (r - q + sigma**2/2)*t)/(sigma*sqrt(t))

        d2 = d1 - sigma*sqrt(t)
    """
    r, q = np.asarray(r), np.asarray(q)
    d1, d2 = bsd1d2(k, t, x0=x0, r=r, q=q, sigma=sigma)
    return x0 * exp(-q*t) * scipy.stats.norm.cdf(d1) - \
        k * exp(-r*t) * scipy.stats.norm.cdf(d2)


# @kfunc(nvar=2)
def bscall_delta(k, t, *, x0=1., r=0., q=0., sigma=1.):
    """
    bscall_delta(k, t, *, x0=1., r=0., q=0., sigma=1.)

    Black-Scholes call option delta.

    See Also
    --------
    bscall
    """
    r, q = np.asarray(r), np.asarray(q)
    d1, d2 = bsd1d2(k, t, x0=x0, r=r, q=q, sigma=sigma)
    return exp(-q*t) * scipy.stats.norm.cdf(d1)


# @kfunc(nvar=2)
def bsput(k, t, *, x0=1., r=0., q=0., sigma=1.):
    """
    bsput(k, t, *, x0=1., r=0., q=0., sigma=1.)

    Black-Scholes put option value.

    See Also
    --------
    bscall
    """
    r, q = np.asarray(r), np.asarray(q)
    d1, d2 = bsd1d2(k, t, x0=x0, r=r, q=q, sigma=sigma)
    return k * exp(-r*t) * scipy.stats.norm.cdf(-d2) - \
        x0 * exp(-q*t) * scipy.stats.norm.cdf(-d1)


# @kfunc(nvar=2)
def bsput_delta(k, t, *, x0=1., r=0., q=0., sigma=1.):
    """
    bsput_delta(k, t, *, x0=1., r=0., q=0., sigma=1.)

    Black-Scholes put option delta.

    See Also
    --------
    bscall
    """
    r, q = np.asarray(r), np.asarray(q)
    d1, d2 = bsd1d2(k, t, x0=x0, r=r, q=q, sigma=sigma)
    return -exp(-q*t) * scipy.stats.norm.cdf(-d1)



import os
import setuptools
from sdepy import __version__ as version

# project root directory
DIR = os.path.dirname(__file__)

# get readme file
with open(os.path.join(DIR, 'README.rst')) as f:
    readme = f.read()

if __name__ == '__main__':
    setuptools.setup(
        name='sdepy',
        version=version,
        description='SdePy: Numerical Integration of Ito '
        'Stochastic Differential Equations',
        long_description=readme,
        long_description_content_type='text/x-rst',
        url='https://github.com/sdepy/sdepy',
        download_url='https://pypi.org/project/sdepy',
        author='Maurizio Cipollina',
        author_email='sdepydev@gmail.com',
        license='BSD',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
            'Operating System :: OS Independent',
            ],
        platforms=[
            'Windows',
            'Linux',
            'Solaris',
            'Mac OS-X',
            'Unix',
            ],
        test_suite='nose.collector',
        python_requires='>=3.5',
        install_requires=['numpy>=1.11.3', 'scipy>=0.18.1', 'nose>=1.3.7'],
        packages=[
            'sdepy',
            'sdepy.tests',
            ],
        package_data={'sdepy': [
            'tests/cfr/*err_expected.txt',
            ]}
        )
